{
    This file is part of the Web Service Toolkit
    Copyright (c) 2006 by Inoussa OUEDRAOGO

    This file is provide under LGPL licence which can be found at
    http://www.gnu.org/licenses/lgpl.html.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
}

unit type_registry;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Contnrs, TypInfo;

Type

  ETypeRegistryException = class(Exception)
  End;

  { TTypeRegistryItem }

  TTypeRegistryItem = class
  private
    FDataType: PTypeInfo;
    FNameSpace: String;
    FDeclaredName : String;
  Public
    constructor Create(
      ANameSpace : String;
      ADataType : PTypeInfo;
      Const ADeclaredName : String = ''
    );
    Property DataType : PTypeInfo Read FDataType;
    Property NameSpace : String Read FNameSpace;
    Property DeclaredName : String Read FDeclaredName;
  End;

  { TTypeRegistry }

  TTypeRegistry = class
  Private
    FList : TObjectList;
    function GetCount: Integer;
    function GetItemByIndex(Index: Integer): TTypeRegistryItem;
    function GetItemByTypeInfo(Index: PTypeInfo): TTypeRegistryItem;
  Public
    constructor Create();
    destructor Destroy();override;
    function IndexOf(Const ATypeInfo : PTypeInfo):Integer;
    function Add(AItem:TTypeRegistryItem):Integer;
    function Register(
      Const ANameSpace    : String;
      Const ADataType     : PTypeInfo;
      Const ADeclaredName : String = ''
    ):Integer;
    function Find(ATypeInfo : PTypeInfo; Const AExact : Boolean):TTypeRegistryItem;
    Property Count : Integer Read GetCount;
    Property Item[Index:Integer] : TTypeRegistryItem Read GetItemByIndex;default;
    Property ItemByTypeInfo[Index:PTypeInfo] : TTypeRegistryItem Read GetItemByTypeInfo;
  End;


  function GetTypeRegistry():TTypeRegistry;
  
implementation
Var
  TypeRegistryInstance : TTypeRegistry = Nil;

function GetTypeRegistry():TTypeRegistry;
begin
  If Not Assigned(TypeRegistryInstance) Then
    TypeRegistryInstance := TTypeRegistry.Create();
  Result := TypeRegistryInstance;
end;

{ TTypeRegistry }

function TTypeRegistry.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TTypeRegistry.GetItemByIndex(Index: Integer): TTypeRegistryItem;
begin
  Result := FList[Index] as TTypeRegistryItem;
end;

function TTypeRegistry.GetItemByTypeInfo(Index: PTypeInfo): TTypeRegistryItem;
begin
  Result := FList[IndexOf(Index)] as TTypeRegistryItem;
end;

constructor TTypeRegistry.Create();
begin
  Inherited Create();
  FList := TObjectList.Create(True);
end;

destructor TTypeRegistry.Destroy();
begin
  FreeAndNil(FList);
  inherited Destroy();
end;

function TTypeRegistry.IndexOf(Const ATypeInfo: PTypeInfo): Integer;
begin
  For Result := 0 To Pred(Count) Do Begin
    If AnsiSameText(ATypeInfo^.Name,Item[Result].DataType^.Name) Then
      Exit;
  End;
  Result := -1;
end;

function TTypeRegistry.Add(AItem: TTypeRegistryItem): Integer;
begin
  Result := IndexOf(AItem.DataType);
  If ( Result = -1 ) Then
    Result := FList.Add(AItem)
  Else
    Raise ETypeRegistryException.CreateFmt('Type already registered : "%s"',[AItem.DataType^.Name]);
end;

function TTypeRegistry.Register(
  Const ANameSpace    : String;
  Const ADataType     : PTypeInfo;
  Const ADeclaredName : String = ''
): Integer;
begin
  Result := IndexOf(ADataType);
  If ( Result = -1 ) Then
    Result := Add(TTypeRegistryItem.Create(ANameSpace,ADataType,ADeclaredName));
end;

function TTypeRegistry.Find(ATypeInfo : PTypeInfo; Const AExact : Boolean):TTypeRegistryItem;
Var
  i : Integer;
  searchClass : TClass;
begin
  Result := Nil;
  i := IndexOf(ATypeInfo);
  If ( i > -1 ) Then
    Result := Item[i]
  Else If ( Not AExact ) And
          Assigned(ATypeInfo) And
          ( ATypeInfo^.Kind = tkClass )
  Then Begin
    searchClass := GetTypeData(ATypeInfo)^.ClassType;
    For i := 0 To Pred(Count) Do Begin
      Result := Item[i];
      If ( Result.DataType^.Kind = tkClass ) And
         searchClass.InheritsFrom(GetTypeData(Result.DataType)^.ClassType)
      Then
        Exit;
    End;
    Result := Nil;
  End;
end;

{ TTypeRegistryItem }

constructor TTypeRegistryItem.Create(
        ANameSpace    : String;
        ADataType     : PTypeInfo;
  Const ADeclaredName : String
);
begin
  FNameSpace := ANameSpace;
  FDeclaredName := ADeclaredName;
  FDataType  := ADataType;
end;


end.
