unit generator;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,
  parserdefs;
  
Type

  { TGenerator }

  TGenerator = class
  Private
    FIndentCount : Integer;
    FDecStream : TStream;
    FImpStream : TStream;
    FCurrentStream : TStream;
    FSymbolTable: TSymbolTable;
    procedure Indent();
    procedure IncIndent();
    procedure DecIndent();
    procedure Write(AText : String);overload;
    procedure Write(AText : String; Const AArgs : array of const);overload;
    procedure WriteLn(AText : String);overload;
    procedure WriteLn(AText : String; Const AArgs : array of const);overload;
    procedure NewLine();
    
    function GenerateStubClassName(AIntf : TInterfaceDefinition):String;
    
    procedure GenerateUnitHeader();
    procedure GenerateUnitImplementationHeader();
    procedure GenerateUnitImplementationFooter();
    procedure GenerateStubIntf(AIntf : TInterfaceDefinition);
    procedure GenerateStubImp(AIntf : TInterfaceDefinition);
  Public
    constructor Create(ASymTable:TSymbolTable);
    destructor Destroy();override;
    procedure Execute();
    procedure GetData(ADestStream:TStream);
    function GetDestUnitName():string;
    property SymbolTable : TSymbolTable Read FSymbolTable;
  End;

implementation
uses StrUtils, parserutils;

Const sSTUB_BASE_CLASS = 'TBaseProxy';
      //sSTREAM_CLASS    = 'TFormaterStream';
      sSERIALIZER_CLASS  = 'TAbstractFormatter';
      RETURN_PARAM_NAME = 'return';

{ TGenerator }

procedure TGenerator.Indent();
Const INDENT_STR = '  ';
begin
  Write(DupeString(INDENT_STR,FIndentCount));
end;

procedure TGenerator.IncIndent();
begin
  Inc(FIndentCount);
end;

procedure TGenerator.DecIndent();
begin
  Dec(FIndentCount);
end;

procedure TGenerator.Write(AText: String);
Var
  i : Integer;
begin
  Assert(Assigned(FCurrentStream));
  i := Length(AText);
  If ( i > 0 ) Then
    FCurrentStream.Write(AText[1],i);
end;

procedure TGenerator.Write(AText: String; const AArgs: array of const);
begin
  Write(Format(AText,AArgs));
end;

procedure TGenerator.WriteLn(AText: String);
begin
  Write(AText+sNEW_LINE);
end;

procedure TGenerator.WriteLn(AText: String; const AArgs: array of const);
begin
  Write(AText+sNEW_LINE,AArgs);
end;

procedure TGenerator.NewLine();
begin
  WriteLn('');
end;

function TGenerator.GenerateStubClassName(AIntf: TInterfaceDefinition): String;
begin
  Result := AIntf.Name;
  If upCase(Result[1]) = 'I' Then
    Delete(Result,1,1);
  Result := Format('T%s_Proxy',[Result]);
end;

procedure TGenerator.GenerateUnitHeader();
begin
  FCurrentStream := FDecStream;
  WriteLn('Unit %s;',[GetDestUnitName()]);
  WriteLn('{$mode objfpc}{$H+}');
  WriteLn('Interface');
  WriteLn('');
  WriteLn('Uses SysUtils, Classes, formaterintf, basestubimp, type_registry, %s;',[SymbolTable.Name]);
  WriteLn('');
  WriteLn('Type');
  WriteLn('');
end;

procedure TGenerator.GenerateUnitImplementationHeader();
begin
  FCurrentStream := FImpStream;
  WriteLn('');
  WriteLn('Implementation');
  WriteLn('');
end;

procedure TGenerator.GenerateUnitImplementationFooter();
begin
  FCurrentStream := FImpStream;
  WriteLn('');
  WriteLn('End.');
end;

constructor TGenerator.Create(ASymTable:TSymbolTable);
begin
  Assert(Assigned(ASymTable));
  FIndentCount := 0;
  FCurrentStream := Nil;
  FSymbolTable := ASymTable;
  FDecStream := TMemoryStream.Create();
  FImpStream := TMemoryStream.Create();
end;

destructor TGenerator.Destroy();
begin
  FCurrentStream := Nil;
  FDecStream.Free();
  FImpStream.Free();
  inherited Destroy();
end;

procedure TGenerator.Execute();
Var
  i,c : Integer;
  intf : TInterfaceDefinition;
begin
  GenerateUnitHeader();
  GenerateUnitImplementationHeader();
  c := Pred(SymbolTable.Count);
  For i := 0 To c Do Begin
    If SymbolTable.Item[i] Is TInterfaceDefinition Then Begin
      intf := SymbolTable.Item[i] As TInterfaceDefinition;
      GenerateStubIntf(intf);
      GenerateStubImp(intf);
    End;
  End;
  GenerateUnitImplementationFooter();
end;

procedure TGenerator.GetData(ADestStream: TStream);
begin
  FDecStream.Position := 0;
  ADestStream.CopyFrom(FDecStream,0);
  FImpStream.Position := 0;
  ADestStream.CopyFrom(FImpStream,0);
end;

function TGenerator.GetDestUnitName(): string;
begin
  Result := Format('%simpunit',[SymbolTable.Name]);
end;

procedure TGenerator.GenerateStubIntf(AIntf: TInterfaceDefinition);
  procedure WriteDec();
  begin
    Indent();
    WriteLn('%s=class(%s,%s)',[GenerateStubClassName(AIntf),sSTUB_BASE_CLASS,AIntf.Name]);
  end;
  
  procedure WriteMethod(AMthd : TMethodDefinition);
  Var
    prmCnt,k : Integer;
    prm : TParameterDefinition;
  Begin
    Indent();
    prmCnt := AMthd.ParameterCount;
    If ( AMthd.MethodType = mtProcedure ) Then
      Write('procedure ')
    Else Begin
      Write('function ');
      Dec(prmCnt);
    End;
    Write('%s(',[AMthd.Name]);

    If ( prmCnt > 0 ) Then Begin
      IncIndent();
      For k := 0 To Pred(prmCnt) Do Begin
        prm := AMthd.Parameter[k];
        If (k > 0 ) Then
          Write('; ');
        NewLine();
        Indent();
        Write('%s %s : %s',[ParameterModifierMAP[prm.Modifier],prm.Name,prm.DataType.Name]);
      End;
      DecIndent();
      NewLine();
      Indent();
    End;

    Write(')');
    If ( AMthd.MethodType = mtFunction ) Then Begin
      Write(':%s',[AMthd.Parameter[prmCnt].DataType.Name]);
    End;
    WriteLn(';');
  End;
  
  procedure WriteMethods();
  Var
    k : Integer;
  begin
    If ( AIntf.MethodCount = 0 ) Then
      Exit;
    IncIndent();
      Indent();
      WriteLn('Protected');
      IncIndent();
        For k := 0 To Pred(AIntf.MethodCount) Do
          WriteMethod(AIntf.Method[k]);
      DecIndent();
    DecIndent();
  end;
begin
  FCurrentStream := FDecStream;
  NewLine();
  IncIndent();
    WriteDec();
    WriteMethods();
    Indent(); WriteLn('End;');
  DecIndent();
end;

procedure TGenerator.GenerateStubImp(AIntf: TInterfaceDefinition);
Const sPRM_NAME = 'strPrmName';
Var
  strClassName : String;
  
  procedure WriteDec();
  begin
    If ( AIntf.MethodCount > 0 ) Then
      WriteLn('{ %s implementation }',[strClassName]);
  end;

  procedure WriteMethodDec(AMthd : TMethodDefinition);
  Var
    prmCnt,k : Integer;
    prm : TParameterDefinition;
  Begin
    prmCnt := AMthd.ParameterCount;
    If ( AMthd.MethodType = mtProcedure ) Then
      Write('procedure ')
    Else Begin
      Write('function ');
      Dec(prmCnt);
    End;
    Write('%s.%s(',[strClassName,AMthd.Name]);

    If ( prmCnt > 0 ) Then Begin
      IncIndent();
      For k := 0 To Pred(prmCnt) Do Begin
        prm := AMthd.Parameter[k];
        If (k > 0 ) Then
          Write('; ');
        NewLine();
        Indent();
        Write('%s %s : %s',[ParameterModifierMAP[prm.Modifier],prm.Name,prm.DataType.Name]);
      End;
      DecIndent();
      NewLine();
      Indent();
    End;

    Write(')');
    If ( AMthd.MethodType = mtFunction ) Then Begin
      Write(':%s',[AMthd.Parameter[prmCnt].DataType.Name]);
    End;
    WriteLn(';');
  End;

  procedure WriteMethodImp(AMthd : TMethodDefinition);
  const sLOC_SERIALIZER = 'locSerializer';
  Var
    prmCnt,k : Integer;
    prm : TParameterDefinition;
  Begin
    IncIndent();
    WriteLn('Var');

      Indent();WriteLn('%s : %s;',[sLOC_SERIALIZER,sSERIALIZER_CLASS]);
      Indent();WriteLn('%s : %s;',[sPRM_NAME,'string']);
    WriteLn('Begin');
    
      Indent();WriteLn('%s := CreateSerializer();',[sLOC_SERIALIZER]);
      Indent();WriteLn('Try');IncIndent();

      Indent();WriteLn('%s.BeginCall(''%s'', GetTarget(),GetAdress());',[sLOC_SERIALIZER,AMthd.Name]);
      IncIndent();
        prmCnt := AMthd.ParameterCount;
        If ( AMthd.MethodType = mtFunction ) Then
          Dec(prmCnt);
        For k := 0 To Pred(prmCnt) Do Begin
          prm := AMthd.Parameter[k];
          If ( prm.Modifier <> pmOut ) Then Begin
            Indent();WriteLn('%s.Put(''%s'', TypeInfo(%s), %s);',[sLOC_SERIALIZER,prm.Name,prm.DataType.Name,prm.Name]);
          End;
        End;
      DecIndent();
      Indent();WriteLn('%s.EndCall();',[sLOC_SERIALIZER]);
      
      WriteLn('');
      Indent();WriteLn('MakeCall(%s);',[sLOC_SERIALIZER]);
      WriteLn('');
      
      Indent();WriteLn('%s.BeginCallRead();',[sLOC_SERIALIZER]);
      IncIndent();
        k:= Pred(AMthd.ParameterCount);
        If ( AMthd.MethodType = mtFunction ) Then Begin
          prm := AMthd.Parameter[k];
          Indent();WriteLn('%s := %s;',[sPRM_NAME,QuotedStr(RETURN_PARAM_NAME)]);
          Indent();WriteLn('%s.Get(TypeInfo(%s), %s, %s);',[sLOC_SERIALIZER,prm.DataType.Name,sPRM_NAME,prm.Name]);
        End;
        For k := 0 To Pred(prmCnt) Do Begin
          prm := AMthd.Parameter[k];
          If ( prm.Modifier In [pmVar, pmOut] ) Then Begin
            Indent();WriteLn('%s := %s;',[sPRM_NAME,QuotedStr(prm.Name)]);
            Indent();WriteLn('%s.Get(TypeInfo(%s), %s, %s);',[sLOC_SERIALIZER,prm.DataType.Name,sPRM_NAME,prm.Name]);
          End;
        End;
      DecIndent();

      
      WriteLn('');
      DecIndent();
      Indent();WriteLn('Finally');
        IncIndent();
          Indent();WriteLn('%s.Free();',[sLOC_SERIALIZER]);
        DecIndent();
      Indent();WriteLn('End;');DecIndent();
      
    WriteLn('End;');
  End;

  procedure WriteMethods();
  Var
    k : Integer;
  begin
    For k := 0 To Pred(AIntf.MethodCount) Do Begin
      WriteMethodDec(AIntf.Method[k]);
      WriteMethodImp(AIntf.Method[k]);
      WriteLn('');
    End;
  end;
  
begin
  FCurrentStream := FImpStream;
  FIndentCount := 0;
  strClassName := GenerateStubClassName(AIntf);
  NewLine();
  WriteDec();
  WriteMethods();
end;

end.
