unit ws_parser;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, parserdefs;

Type
  EParserException = class(Exception)
  End;
  
  { TPascalParser }

  TPascalParser = class
  private
    FStream : TStream;
    FTokenizer : TParser;
    FErrorMessage : String;
    FSymbolTable : TSymbolTable;
    FCurrentSymbol : TAbstractSymbolDefinition;
  private
    Property Tokenizer : TParser Read FTokenizer;
    procedure SetErrorMessage(Const AValue : String);
    procedure CheckCurrentString(Const AString : String);
  private
    procedure BeginParsing();
    procedure EndParsing();
    function GetSourceLine: Integer;
    function ReadUntil(Const AText : String; Const ARaiseException : Boolean):Boolean;
  Private
    procedure ParseHeader();
    procedure ParseInterfaceSection();
    procedure ParseUses();
    procedure ParseTypeDeclaration();
    procedure ParseInterfaceType(Const AName : String);
    procedure ParseEnumType(Const AName : String);
  public
    constructor Create(AStream : TStream);
    destructor Destroy();override;
    procedure Error(Const AMsg : String);overload;
    procedure Error(Const AMsg : String; Const AArgs : Array of const);overload;
    function Parse():Boolean;
    property SourceLine: Integer read GetSourceLine;
    property ErrorMessage : String read FErrorMessage;
    property SymbolTable : TSymbolTable Read FSymbolTable;
  End;
  
  
implementation

Type
  TPascalToken = (
    ptNone,
    ptUnit, ptInterface, ptUses, ptType,
    ptImplementation, ptEnd,
    ptProcedure, ptFunction,
    ptSemicolon, ptComma, ptPeriod, ptEqual, ptColon,
    ptLeftParenthesis, ptRigthParenthesis,
    ptConst, ptVar, ptOut
  );
Const
  PascalTokenStrMAP : Array[TPascalToken] Of String = (
    '',
    'UNIT', 'INTERFACE', 'USES', 'TYPE',
    'IMPLEMENTATION', 'END',
    'PROCEDURE', 'FUNCTION',
    ';', ',', '.', '=', ':',
    '(', ')',
    'CONST', 'VAR', 'OUT'
  );
  
function GetPascalTokenStr(Const AToken:TPascalToken):String;
begin
  Result := PascalTokenStrMAP[AToken];
end;

function GetPascalTokenFromStr(Const ATokenStr:String):TPascalToken;
begin
  Result := Succ(ptNone);
  For Result := Result To High(TPascalToken) Do Begin
    If AnsiSameText(ATokenStr,PascalTokenStrMAP[Result]) Then
      Exit;
  End;
  Result := ptNone;
end;

{ TPascalParser }

procedure TPascalParser.SetErrorMessage(const AValue: String);
begin
  FErrorMessage := AValue;
end;

procedure TPascalParser.Error(const AMsg: String);
begin
  Raise EParserException.Create(AMsg);
end;

procedure TPascalParser.Error(const AMsg: String; const AArgs: array of const);
begin
  Raise EParserException.CreateFmt(AMsg,AArgs);
end;

procedure TPascalParser.CheckCurrentString(const AString: String);
begin
  If Not AnsiSameText(Tokenizer.TokenString,AString) Then
    Error('"%s" expected.',[AString]);
end;

procedure TPascalParser.BeginParsing();
begin

end;

procedure TPascalParser.EndParsing();
begin

end;

function TPascalParser.GetSourceLine: Integer;
begin
  Result := Tokenizer.SourceLine;
end;

function TPascalParser.ReadUntil(Const AText : String; Const ARaiseException : Boolean):Boolean;
begin
  While ( Tokenizer.Token <> toEOF ) And ( Not AnsiSameText(Tokenizer.TokenString,AText) ) Do
    Tokenizer.NextToken();
  Result := AnsiSameText(Tokenizer.TokenString,AText);
  If ARaiseException And ( Not Result ) Then
    Error('"%s" non trouvé.',[AText]);
end;

procedure TPascalParser.ParseHeader();
begin // Unit UnitName;
  Tokenizer.CheckTokenSymbol(GetPascalTokenStr(ptUnit));
  Tokenizer.NextToken();
  Tokenizer.CheckToken(toSymbol);
  FSymbolTable.Name := Tokenizer.TokenString;
  Tokenizer.NextToken();
  CheckCurrentString(GetPascalTokenStr(ptSemicolon));

  Tokenizer.NextToken();
end;

procedure TPascalParser.ParseInterfaceSection();
begin
  ReadUntil(GetPascalTokenStr(ptInterface),True);
  Tokenizer.CheckTokenSymbol(GetPascalTokenStr(ptInterface));
  Tokenizer.NextToken();
  Tokenizer.CheckToken(toSymbol);
  If Tokenizer.TokenSymbolIs(GetPascalTokenStr(ptUses)) Then
    ParseUses();
  Tokenizer.CheckToken(toSymbol);
  If Tokenizer.TokenSymbolIs(GetPascalTokenStr(ptType)) Then Begin
    Tokenizer.NextToken();
    Repeat
      ParseTypeDeclaration();
    Until ( Tokenizer.Token = toEOF ) Or ( Tokenizer.TokenSymbolIs(GetPascalTokenStr(ptImplementation)) );
  End;
end;

procedure TPascalParser.ParseUses();
begin
  Tokenizer.CheckTokenSymbol(GetPascalTokenStr(ptUses));
  Tokenizer.NextToken();
  Repeat
    Tokenizer.CheckToken(toSymbol); //UnitName
    Tokenizer.NextToken();
    If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptSemicolon)) Then
      Break;
    CheckCurrentString(GetPascalTokenStr(ptComma));
    Tokenizer.NextToken();
  Until ( Tokenizer.Token = toEOF ) Or AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptSemicolon));
  CheckCurrentString(GetPascalTokenStr(ptSemicolon));
  Tokenizer.NextToken();
end;

procedure TPascalParser.ParseTypeDeclaration();
Var
  sname : String;
begin
  Tokenizer.CheckToken(toSymbol);
  sname := Tokenizer.TokenString;
  Tokenizer.NextToken();
  CheckCurrentString(GetPascalTokenStr(ptEqual));

  Tokenizer.NextToken();
  If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptLeftParenthesis)) Then
    self.ParseEnumType(sname)
  Else Begin
    Tokenizer.CheckToken(toSymbol);
    If Tokenizer.TokenSymbolIs(GetPascalTokenStr(ptInterface)) Then
      ParseInterfaceType(sname)
    Else Begin
      ReadUntil(GetPascalTokenStr(ptEnd),True);
      Tokenizer.NextToken();// End
      Tokenizer.NextToken();// ;
    End;
  End;
end;

procedure TPascalParser.ParseInterfaceType(const AName: String);
Var
  sbl : TInterfaceDefinition;
  
  procedure ReadIntfHeader();
  Var
    tmpStr : String;
  begin
    Tokenizer.NextToken();
    Repeat
      Tokenizer.CheckToken(toSymbol);
      tmpStr := Tokenizer.TokenString;
      Tokenizer.NextToken();
      If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptRigthParenthesis)) Then Begin
        Tokenizer.NextToken();
        Break;
      End;
      Tokenizer.NextToken();
      CheckCurrentString(GetPascalTokenStr(ptComma));

      Tokenizer.NextToken();
    Until ( Tokenizer.Token = toEOF ) ;
  end;
  
  procedure ReadProcedure(Const AProc : Boolean);
  Var
    tmpStr,prmName : String;
    tmpTkn : TPascalToken;
    pr : TMethodDefinition;
    prmM : TParameterModifier;
    dataTypeSbl : TTypeDefinition;
    foundSymbol : TAbstractSymbolDefinition;
  begin
    Tokenizer.NextToken();
    Tokenizer.CheckToken(toSymbol);
    tmpStr := Tokenizer.TokenString;
    If AProc Then
      pr := sbl.AddMethod(tmpStr,mtProcedure)
    Else
      pr := sbl.AddMethod(tmpStr,mtFunction);
    Tokenizer.NextToken();
    If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptLeftParenthesis)) Then Begin
      Tokenizer.NextToken();
      If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptRigthParenthesis)) Then
        Tokenizer.NextToken()
      Else Begin
        Repeat
          Tokenizer.CheckToken(toSymbol);
          tmpStr := Tokenizer.TokenString;
          tmpTkn := GetPascalTokenFromStr(tmpStr);
          prmM := pmNone;
          If ( tmpTkn = ptConst ) Then Begin
            prmM := pmConst;
          End Else If ( tmpTkn = ptVar ) Then Begin
            prmM := pmVar;
          End Else If ( tmpTkn = ptOut ) Then Begin
            prmM := pmOut;
          End;
          If ( prmM > pmNone ) Then Begin
            Tokenizer.NextToken();
            tmpStr := Tokenizer.TokenString;
          End;
          prmName := tmpStr;
          Tokenizer.NextToken();
          CheckCurrentString(GetPascalTokenStr(ptColon));

          Tokenizer.NextToken();
          Tokenizer.CheckToken(toSymbol);
          tmpStr := Tokenizer.TokenString;
          foundSymbol := FSymbolTable.Find(tmpStr);
          If Assigned(foundSymbol) And ( Not foundSymbol.InheritsFrom(TTypeDefinition) ) Then
            Error('Type symbol expected where "%s" was found.',[foundSymbol.Name]);
          If Assigned(foundSymbol) Then
            dataTypeSbl := foundSymbol As TTypeDefinition
          Else Begin
            dataTypeSbl := TTypeDefinition.Create(tmpStr);
            FSymbolTable.Add(dataTypeSbl);
          End;
          pr.AddParameter(prmName,prmM,dataTypeSbl);
          Tokenizer.NextToken();
          tmpStr := Tokenizer.TokenString;
          If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptRigthParenthesis)) Then Begin
            Tokenizer.NextToken();
            Break;
          End;
          Tokenizer.NextToken();
        Until ( Tokenizer.Token = toEOF ) ;
      End;
    End;
    If Not AProc Then Begin
      CheckCurrentString(GetPascalTokenStr(ptColon));
      Tokenizer.NextToken();
      Tokenizer.CheckToken(toSymbol);
      tmpStr := Tokenizer.TokenString;
      dataTypeSbl := FSymbolTable.Find(tmpStr) As TTypeDefinition;
      If Not Assigned(dataTypeSbl) Then Begin
        dataTypeSbl := TTypeDefinition.Create(tmpStr);
        FSymbolTable.Add(dataTypeSbl);
      End;
      pr.AddParameter('result',pmOut,dataTypeSbl);
      Tokenizer.NextToken();
    End;
    CheckCurrentString(GetPascalTokenStr(ptSemicolon));
    Tokenizer.NextToken();
  end;
  
  procedure ReadFunction();
  begin
    ReadProcedure(False);
  end;
  
begin
  Tokenizer.CheckTokenSymbol(GetPascalTokenStr(ptInterface));
  sbl := TInterfaceDefinition.Create(AName);
  FSymbolTable.Add(sbl);
  FCurrentSymbol := sbl;
  Tokenizer.NextToken();
  If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptLeftParenthesis)) Then
    ReadIntfHeader();
  Repeat
    If Tokenizer.TokenSymbolIs(GetPascalTokenStr(ptProcedure)) Then
      ReadProcedure(True)
    Else If Tokenizer.TokenSymbolIs(GetPascalTokenStr(ptFunction)) Then
      ReadFunction()
    Else If Tokenizer.TokenSymbolIs(GetPascalTokenStr(ptEnd)) Then Begin
      Tokenizer.NextToken();
      Tokenizer.NextToken();
      Break;
    End Else Begin
      Error('"%s", "%s", "%s" attendu.',[GetPascalTokenStr(ptProcedure),GetPascalTokenStr(ptFunction),GetPascalTokenStr(ptEnd)]);
    End;
  Until ( Tokenizer.Token = toEOF ) ;
end;

procedure TPascalParser.ParseEnumType(const AName: String);
Var
  sbl : TEnumTypeDefinition;
  tmpStr : String;
  sblItem : TEnumItemDefinition;
  tmpInt : Integer;
begin
  sbl := TEnumTypeDefinition.Create(AName);
  FSymbolTable.Add(sbl);
  FCurrentSymbol := sbl;
  CheckCurrentString(GetPascalTokenStr(ptLeftParenthesis));
  Tokenizer.NextToken();
  Tokenizer.CheckToken(toSymbol);
  tmpInt := 0;
  Repeat
    tmpStr := Tokenizer.TokenString;
    If ( FSymbolTable.IndexOf(tmpStr) > -1 ) Then
      Error('Duplicated symbol : "%s"',[tmpStr]);
    sblItem := TEnumItemDefinition.Create(tmpStr,tmpInt);
    FSymbolTable.Add(sblItem);
    sbl.AddItem(sblItem);
    Tokenizer.NextToken();
    If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptRigthParenthesis)) Then
      Break;
    If AnsiSameText(Tokenizer.TokenString,GetPascalTokenStr(ptComma)) Then
      Tokenizer.NextToken();
    Tokenizer.CheckToken(toSymbol);
    Inc(tmpInt);
  Until ( Tokenizer.Token = toEOF ) ;
  CheckCurrentString(GetPascalTokenStr(ptRigthParenthesis));
  Tokenizer.NextToken();
  CheckCurrentString(GetPascalTokenStr(ptSemicolon));
  Tokenizer.NextToken();
end;

constructor TPascalParser.Create(AStream : TStream);
begin
  Assert(Assigned(AStream));
  FStream := AStream;
  FTokenizer := TParser.Create(FStream);
  FSymbolTable := TSymbolTable.Create('');
  FCurrentSymbol := Nil;
end;

destructor TPascalParser.Destroy();
begin
  FTokenizer.Free();
  FreeAndNil(FSymbolTable);
  inherited Destroy();
end;

function TPascalParser.Parse(): Boolean;
begin
  BeginParsing();
  Try
    Try
      ParseHeader();
      ParseInterfaceSection();
    Except
      On E : Exception Do Begin
        Result := False;
        SetErrorMessage(E.Message );
      End;
    End;
  Finally
    EndParsing();
  End;
end;

end.

